// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <be/interface/Bitmap.h>
#include <be/interface/Window.h>

#include <be/media/MediaRoster.h>

#include <posix/string.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "Connection.h"
#include "EavesdropperMain.h"
#include "IconFactory.h"
#include "LiveNodesView.h"
#include "LiveNodeView.h"
#include "ConnectorView.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

/**
 * The minimum number of pixels the mouse must have moved inside the view
 * before we start dragging it.
 */
static const int32 DRAG_THRESHOLD = 4;

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

LiveNodeView::LiveNodeView(
		InfoWindow* infoWindow,
		LiveNodesView* liveNodesView,
		live_node_info liveNodeInfo)
	: AbstractNodeView(infoWindow, liveNodeInfo.name)
{
	// Initialize attributes.
	this->liveNodesView = liveNodesView;
	this->liveNodeInfo  = liveNodeInfo;

	// Get interesting stuff about the node from the Media Server.
	GetNodeInfo();
	
	BBG_WRITE(("[LiveNodeView::LiveNodeView] Constructed node %s.",
		liveNodeInfo.name));
}

// ===========================================================================
// Destructor
// ===========================================================================

LiveNodeView::~LiveNodeView(void)
{
	// Delete the list of input connectors.
	for (int32 t=0; t<inputConnectors->CountItems(); ++t)
	{
		delete inputConnectors->ItemAt(t);
	}
	delete inputConnectors;

	// Delete the list of output connectors.
	for (int32 t=0; t<outputConnectors->CountItems(); ++t)
	{
		delete outputConnectors->ItemAt(t);
	}
	delete outputConnectors;

	// Delete the list of file formats.
	for (int32 t=0; t<fileFormats->CountItems(); ++t)
	{
		delete fileFormats->ItemAt(t);
	}
	delete fileFormats;
	
	// Delete the list of attributes.
	for (int32 t=0; t<attributes->CountItems(); ++t)
	{
		delete attributes->ItemAt(t);
	}
	delete attributes;
}

// ===========================================================================
// AttachedToWindow
// ===========================================================================

void
LiveNodeView::AttachedToWindow(void)
{
	// Add all input connectors to the live nodes view.
	if (inputConnectors != 0)
	{
		int32 countConnectors =	inputConnectors->CountItems();
		
		float start = (fontSize + 61 - countConnectors*CONNECTOR_HEIGHT)/2.0
			- liveNodesView->LeftTop().y;
		
		for (int32 t=0; t<countConnectors; ++t)
		{
			BView* connectorView = (BView*) inputConnectors->ItemAt(t);
			
			liveNodesView->AddChild(connectorView);

			connectorView->MoveTo(-CONNECTOR_WIDTH,
				start+t*CONNECTOR_HEIGHT);
		}
	}

	// Add all output connectors to the live nodes view.
	if (outputConnectors != 0)
	{
		int32 countConnectors = outputConnectors->CountItems();

		float start = (fontSize + 61 - countConnectors*CONNECTOR_HEIGHT)/2.0
			- liveNodesView->LeftTop().y;
								
		for (int32 t=0; t<countConnectors; ++t)
		{
			BView* connectorView = (BView*) outputConnectors->ItemAt(t);
			
			liveNodesView->AddChild(connectorView);
			
			connectorView->MoveTo(Bounds().Width()+1,
				start+t*CONNECTOR_HEIGHT);
		}
	}
}

// ===========================================================================
// DetachedFromWindow
// ===========================================================================

void
LiveNodeView::DetachedFromWindow(void)
{
	// Remove all input connectors from the live nodes view.
	if (inputConnectors != 0)
	{
		for (int32 t=0; t<inputConnectors->CountItems(); ++t)
		{
			BView* connectorView = (BView*) inputConnectors->ItemAt(t);
			connectorView->RemoveSelf();
		}
	}

	// Remove all output connectors from the live nodes view.
	if (outputConnectors != 0)
	{
		for (int32 t=0; t<outputConnectors->CountItems(); ++t)
		{
			BView* connectorView = (BView*) outputConnectors->ItemAt(t);
			connectorView->RemoveSelf();
		}
	}
}

// ===========================================================================
// MoveTo
// ===========================================================================

void
LiveNodeView::MoveTo(BPoint point)
{
	float deltaX = point.x - Frame().left;
	float deltaY = point.y - Frame().top;

	// Tell all the input connectors to move.
	if (inputConnectors != 0)
	{
		for (int32 t=0; t<inputConnectors->CountItems(); ++t)
		{
			ConnectorView* connectorView =
				(ConnectorView*) inputConnectors->ItemAt(t);

			connectorView->MoveBy(deltaX, deltaY);
		}
	}

	// Tell all the output connectors to move.
	if (outputConnectors != 0)
	{
		for (int32 t=0; t<outputConnectors->CountItems(); ++t)
		{
			ConnectorView* connectorView =
				(ConnectorView*) outputConnectors->ItemAt(t);

			connectorView->MoveBy(deltaX, deltaY);
		}
	}

	BView::MoveTo(point);
}

// ===========================================================================
// Draw
// ===========================================================================

void 
LiveNodeView::Draw(BRect updateRect)
{
	BView::Draw(updateRect);
	DrawNode(liveNodeInfo.node.node, liveNodeInfo.node.kind);
}

// ===========================================================================
// MouseDown
// ===========================================================================

void
LiveNodeView::MouseDown(BPoint where)
{
	AbstractNodeView::MouseDown(where);

	// Retrieve the button state from the MouseDown message.
	uint32 buttons;
	if (Window()->CurrentMessage()->FindInt32("buttons", (int32*) &buttons)
		== B_NO_ERROR)
	{
		// Did the user click the primary mouse button?
		if (buttons & B_PRIMARY_MOUSE_BUTTON)
		{
			// Start tracking to see if the user wants to drag the view.
			isTracking = true;
			clickPoint = where;
		}
	}
}

// ===========================================================================
// MouseUp
// ===========================================================================

void 
LiveNodeView::MouseUp(BPoint where)
{ 
	// If the mouse button is released, then we must stop tracking the mouse.
	isTracking = false;
}

// ===========================================================================
// MouseMoved
// ===========================================================================

void
LiveNodeView::MouseMoved(BPoint point,
	uint32 transit,
	const BMessage* message)
{
	AbstractNodeView::MouseMoved(point, transit, message);
	
	// Are we currently tracking the mouse?
	if (isTracking)
	{
		// Check to see if mouse has moved more than DRAG_THRESHOLD pixels
		// in any direction.
		if ((abs((int)(clickPoint.x - point.x)) > DRAG_THRESHOLD) 
		||  (abs((int)(clickPoint.y - point.y)) > DRAG_THRESHOLD))
		{
			// Start the drag-and-drop session. 
			BMessage message = new BMessage(MSG_DRAG_NODE);
			message.AddPointer("LiveNodeView", this);
			message.AddPoint("Offset", clickPoint);
	
			DragMessage(&message, Bounds());

			isTracking = false;
		}
	}
}

// ===========================================================================
// GetLiveNodeInfo
// ===========================================================================

live_node_info
LiveNodeView::GetLiveNodeInfo(void)
{
	return liveNodeInfo;
}
	
// ===========================================================================
// GetNode
// ===========================================================================

media_node
LiveNodeView::GetNode(void)
{
	return liveNodeInfo.node;
}
	
// ===========================================================================
// GetInputConnectors
// ===========================================================================

BList*
LiveNodeView::GetInputConnectors(void)
{
	return inputConnectors;
}

// ===========================================================================
// GetOutputConnectors
// ===========================================================================

BList*
LiveNodeView::GetOutputConnectors(void)
{
	return outputConnectors;
}

// ===========================================================================
// GetFileFormats
// ===========================================================================

BList*
LiveNodeView::GetFileFormats(void)
{
	return fileFormats;
}

// ===========================================================================
// GetAttributes
// ===========================================================================

BList*
LiveNodeView::GetAttributes(void)
{
	return attributes;
}

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// GetNodeInfo
// ===========================================================================

void
LiveNodeView::GetNodeInfo(void)
{
	/*!
	 * [TODO] LiveNodeView::GetNodeInfo - Create just enough memory to hold
	 * the inputs, outputs, file formats, and so on.
	 *
	 * At this time of writing, GetFreeInputsFor() and the other routines
	 * do not return the current number of things you're asking for. If this
	 * still isn't fixed in R4.5 final, then implement Owen's workaround. 
	 */

	BMediaRoster* mediaRoster = BMediaRoster::CurrentRoster();

	BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get info for node %s.",
		liveNodeInfo.name));

	inputConnectors = new BList(5);
	outputConnectors = new BList(5);
	fileFormats = new BList(5);
	attributes = new BList(5);
		
	// -----------------------------------------------------------------------
	// Get the connected inputs.
	
	if (liveNodeInfo.node.kind & B_BUFFER_CONSUMER)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get connected"
			" inputs."));
		
		const int32 NUMBER_OF_CONNECTED_INPUTS_TO_TRY = 50;
		int32 numConnectedInputs = 0;

		// Allocate a temporary array to hold the results we get back from
		// the Media Server. Note that we're probably allocating much more
		// memory than we actually need.
		media_input* connectedInputs = (media_input*)
			malloc(sizeof(media_input) * NUMBER_OF_CONNECTED_INPUTS_TO_TRY);

		// Query the Media Server.
		status_t result = mediaRoster->GetConnectedInputsFor(
			liveNodeInfo.node,
			connectedInputs,
			NUMBER_OF_CONNECTED_INPUTS_TO_TRY,
			&numConnectedInputs);

		// Was query successful?
		if (result == B_OK)
		{
			BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld connected inputs.",
				numConnectedInputs));

			for (int32 t=0; t<numConnectedInputs; ++t)
			{
				// Create a new connector for this input. We pass it a
				// pointer to a copy of the media_input structure, using
				// media_input's default copy constructor.
				ConnectorView* connectorView = new ConnectorView(
					infoWindow,
					new media_input(connectedInputs[t]),
					0,
					true);
				
				inputConnectors->AddItem(connectorView);
			}
		}

		// Free the temporary array. We can safely free the malloc()ed memory
		// here, because we're no longer using it. The pointers that have
		// been added to the connectedInputs list reference _copies_ of the
		// media_input structures that we allocated memory for.
		free(connectedInputs);
	}
	
	// -----------------------------------------------------------------------
	// Get the free inputs.
	
	if (liveNodeInfo.node.kind & B_BUFFER_CONSUMER)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get free"
			" inputs."));
		
		const int32 NUMBER_OF_FREE_INPUTS_TO_TRY = 50;
		int32 numFreeInputs = 0;

		// Allocate a temporary array to hold the results we
		// get back from the Media Server.
		media_input* freeInputs = (media_input*)
			malloc(sizeof(media_input) * NUMBER_OF_FREE_INPUTS_TO_TRY);

		// Query the Media Server.
		status_t result = mediaRoster->GetFreeInputsFor(
			liveNodeInfo.node,
			freeInputs,
			NUMBER_OF_FREE_INPUTS_TO_TRY,
			&numFreeInputs);

		// Was query successful?
		if (result == B_OK)
		{
			BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld free inputs.",
				numFreeInputs));

			for (int32 t=0; t<numFreeInputs; ++t)
			{
				ConnectorView* connectorView = new ConnectorView(
					infoWindow,
					new media_input(freeInputs[t]),
					0,
					false);
				
				inputConnectors->AddItem(connectorView);
			}
		}

		free(freeInputs);
	}

	// -----------------------------------------------------------------------
	// Get the connected outputs.
	
	if (liveNodeInfo.node.kind & B_BUFFER_PRODUCER)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get connected"
			" outputs."));
			
		const int32 NUMBER_OF_CONNECTED_OUTPUTS_TO_TRY = 50;
		int32 numConnectedOutputs = 0;

		media_output* connectedOutputs = (media_output*)
			malloc(sizeof(media_output) * NUMBER_OF_CONNECTED_OUTPUTS_TO_TRY);
                          
		status_t result = mediaRoster->GetConnectedOutputsFor(
			liveNodeInfo.node,
			connectedOutputs,
			NUMBER_OF_CONNECTED_OUTPUTS_TO_TRY,
			&numConnectedOutputs);

		if (result == B_OK)
		{
			BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld connected outputs.",
				numConnectedOutputs));

			for (int32 t=0; t<numConnectedOutputs; ++t)
			{
				ConnectorView* connectorView = new ConnectorView(
					infoWindow,
					0,
					new media_output(connectedOutputs[t]),
					true);
				
				outputConnectors->AddItem(connectorView);
			}
		}
		
		free(connectedOutputs);
	}

	// -----------------------------------------------------------------------
	// Get the free outputs.
	
	if (liveNodeInfo.node.kind & B_BUFFER_PRODUCER)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get free"
			" outputs."));
			
		const int32 NUMBER_OF_FREE_OUTPUTS_TO_TRY = 50;
		int32 numFreeOutputs = 0;

		media_output* freeOutputs = (media_output*)
			malloc(sizeof(media_output) * NUMBER_OF_FREE_OUTPUTS_TO_TRY);
                          
		status_t result = mediaRoster->GetFreeOutputsFor(
			liveNodeInfo.node,
			freeOutputs,
			NUMBER_OF_FREE_OUTPUTS_TO_TRY,
			&numFreeOutputs);

		if (result == B_OK)
		{
			BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld free outputs.",
				numFreeOutputs));

			for (int32 t=0; t<numFreeOutputs; ++t)
			{
				ConnectorView* connectorView = new ConnectorView(
					infoWindow,
					0,
					new media_output(freeOutputs[t]),
					false);
				
				outputConnectors->AddItem(connectorView);
			}
		}
		
		free(freeOutputs);
	}

	// -----------------------------------------------------------------------
	// Get the supported file formats.

	if (liveNodeInfo.node.kind & B_FILE_INTERFACE)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get supported"
			" file formats."));

		const int32 NUMBER_OF_FILE_FORMATS_TO_TRY = 50;
		int32 numFileFormats = NUMBER_OF_FILE_FORMATS_TO_TRY;

		media_file_format* temp = (media_file_format*)
			malloc(sizeof(media_file_format) * NUMBER_OF_FILE_FORMATS_TO_TRY);

		status_t result = mediaRoster->GetFileFormatsFor(
			liveNodeInfo.node,
			temp,
			&numFileFormats);

		if (result == B_OK)
		{
	        BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld file formats.",
	                   numFileFormats));

			for (int32 t=0; t<numFileFormats; ++t)
			{
				fileFormats->AddItem(new media_file_format(temp[t]));
			}
		}
		
		free(temp);
	}

	// -----------------------------------------------------------------------
	// Get the node's attributes.

	BBG_WRITE(("[LiveNodeView::GetNodeInfo] Trying to get attributes."));

	const size_t NUMBER_OF_ATTRIBUTES_TO_TRY = 50;

	media_node_attribute* temp = (media_node_attribute*)
		malloc(sizeof(media_node_attribute) * NUMBER_OF_ATTRIBUTES_TO_TRY);
                          
	ssize_t numAttributes = mediaRoster->GetNodeAttributesFor(
		liveNodeInfo.node,
		temp,
		NUMBER_OF_ATTRIBUTES_TO_TRY);
		
	if (numAttributes >= 0)
	{
		BBG_WRITE(("[LiveNodeView::GetNodeInfo] Got %ld attributes.",
			numAttributes));
			
		for (int32 t=0; t<numAttributes; ++t)
		{
			attributes->AddItem(new media_node_attribute(temp[t]));
		}
		
		free (temp);
	}
	         	
	// -----------------------------------------------------------------------
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
